@@ -19,9 +19,13 @@ and volume, and notes are added given a specific duration using those
 values. However, it's not like playing an instrument, where time is
 continuous and irreversible: it's possible to change the current
 time-point to write multiple notes that overlap each other.
+
+A current key is tracked, and you can change the fundamental note and
+select from a few different scale types including major, minor, and
+pentatonic scales.
 """
 
-__version__ = "1.0.3"
+__version__ = "2.0.0"
 
 import math, random, struct, wave
 
@@ -87,7 +91,9 @@ Default relax duration for the limiter, in frames.
 
 SEMITONE = 2 ** (1 / 12)
 
-PITCH_NAMES = ["C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B"]
+PITCH_NAMES = [
+    "C", "Db", "D", "Eb", "E", "F", "Gb", "G", "Ab", "A", "Bb", "B"
+]
 SHARP_NAMES = {
     "Db": "Cs",
     "Eb": "Ds",
@@ -139,6 +145,189 @@ for _name, _freq in PENTATONIC_NOTES:
     globals()[_name] = _freq
 
 
+#----------------#
+# Key management #
+#----------------#
+
+
+def currentKey():
+    """
+    Returns the current key as a pair of fundamental note name plus
+    scale type (see `CURRENT_KEY`).
+    """
+    return CURRENT_KEY
+
+
+def currentFundamentalName():
+    """
+    Returns the name of the current fundamental note, which by default
+    is C0.
+    """
+    return CURRENT_KEY[0] + "0"
+
+
+def currentFundamental():
+    """
+    Returns the numeric pitch value for the fundamental note of the
+    current key. This is based on the key's fundamental note name string
+    in the octave 0. Raises an error if the key string is invalid.
+    """
+    noteName = CURRENT_KEY[0] + "0"
+    g = globals()
+    if noteName not in g:
+        raise ValueError(
+            "Current key's fundamental note '{}' is invalid.".format(
+                noteName
+            )
+        )
+    return g[noteName]
+
+
+def setFundamental(noteName):
+    """
+    Sets the fundamental pitch for the current key, leaving the scale
+    type unchanged. The given noteName must be a valid pitch name, which
+    is one of:
+
+           A    B C    D    E F    G
+        Ab   Bb     Db   Eb     Gb
+             As     Cs   Ds     Fs   Gs
+
+    A trailing 'b' or 's' indicates flat ('b') or sharp ('s'). If a name
+    with 's' is given, it will be converted into the equivalent 'b'
+    name.
+    """
+    global CURRENT_KEY
+
+    if not isinstance(noteName, str):
+        raise TypeError("The note name must be a string.")
+
+    if noteName not in PITCH_NAMES and noteName not in SHARP_NAMES.values():
+        raise ValueError(
+            (
+                "Invalid fundamental note name: '{}'. A note name must"
+                " start with a letter A-G and may be followed by 'b'"
+                " for flat or 's' for sharp; Cb/Bs and Fb/Es are not"
+                " allowed."
+            ).format(noteName)
+        )
+
+    # Convert to canonical form
+    if noteName in SHARP_NAMES.values():
+        # Lazy dictionary inversion
+        for name in SHARP_NAMES:
+            if SHARP_NAMES[name] == noteName:
+                noteName = name
+                break
+
+    CURRENT_KEY = (noteName, CURRENT_KEY[1])
+
+
+def currentScaleType():
+    """
+    Returns a string describing the current scale type, or for custom
+    scales, either a list of semitone intervals or a pair of such lists.
+    """
+    return CURRENT_KEY[1]
+
+
+def ascendingScaleIntervals():
+    """
+    Returns the list of semitone intervals in the current ascending
+    scale.
+    """
+    if isinstance(CURRENT_KEY[1], str):
+        if CURRENT_KEY[1] not in SCALE_TYPES:
+            raise ValueError(
+                (
+                    "Invalid scale type '{}': named scale types must be"
+                    " one of:\n  {}"
+                ).format(
+                    CURRENT_KEY[1],
+                    '\n  '.join(SCALE_TYPES.keys())
+                )
+            )
+        intervals = SCALE_TYPES[CURRENT_KEY[1]]
+    else:
+        intervals = CURRENT_KEY[1]
+
+    if isinstance(intervals, tuple):
+        return intervals[0]
+    elif isinstance(intervals, list):
+        return intervals
+    else:
+        raise TypeError(
+            (
+                "Custom scale types must be lists or tuples (the"
+                " current scale type is a {})."
+            ).format(type(intervals))
+        )
+
+
+def descendingScaleIntervals():
+    """
+    Returns the list of semitone intervals in the current descending
+    scale.
+    """
+    if isinstance(CURRENT_KEY[1], str):
+        if CURRENT_KEY[1] not in SCALE_TYPES:
+            raise ValueError(
+                (
+                    "Invalid scale type '{}': named scale types must be"
+                    " one of:\n  {}"
+                ).format(
+                    CURRENT_KEY[1],
+                    '\n  '.join(SCALE_TYPES.keys())
+                )
+            )
+        intervals = SCALE_TYPES[CURRENT_KEY[1]]
+    else:
+        intervals = CURRENT_KEY[1]
+
+    if isinstance(intervals, tuple):
+        return intervals[1]
+    elif isinstance(intervals, list):
+        return intervals
+    else:
+        raise TypeError(
+            (
+                "Custom scale types must be lists or tuples (the"
+                " current scale type is a {})."
+            ).format(type(intervals))
+        )
+
+
+def setScaleType(scaleType):
+    """
+    Sets the current scale type, leaving the fundamental note for the
+    current key unchanged. The `SCALE_TYPES` variable defines available
+    scale types, and one of the keys from that dictionary should be
+    used. For a custom scale type, a list of integers may be provided
+    instead, defining the semitones between successive notes in the
+    scale. If different ascending/descending note sequences are desired,
+    a pair of such lists may be provided specifying first the ascending
+    and then the descending note sequences.
+    """
+    if not isinstance(scaleType, (list, tuple, str)):
+        raise TypeError(
+            (
+                "The scale type must be a string, list, or tuple (you"
+                " provided a/an {})"
+            ).format(type(scaleType))
+        )
+
+    if isinstance(scaleType, str) and scaleType not in SCALE_TYPES:
+        raise ValueError(
+            (
+                "Invalid scale type '{}': named scale types must be"
+                " one of:\n  {}"
+            ).format(
+                scaleType,
+                '\n  '.join(SCALE_TYPES.keys())
+            )
+        )
+
+
 #-----------------#
 # Tone management #
 #-----------------#
@@ -196,119 +385,89 @@ def pentatonicIndex(pitch):
     return best
 
 
-def stepUpFrom(pitch, nSteps=1):
+def climbUpFrom(pitch, nRungs=1):
     """
-    Returns the tone that's nSteps (default 1) whole-step(s) above the
-    given pitch, on an 8-note octave. Detects the nearest piano-key pitch
-    and gives the tone in steps above (or below if nSteps is negative)
-    that.
+    Returns the tone that's nRungs (default 1) scale-degrees above the
+    given pitch based on the current key (see `CURRENT_KEY`). Detects
+    the nearest in-key pitch and gives the tone nRungs degrees above
+    that pitch.
 
-    If the pitch is out of range of the piano keys, simply returns a
-    pitch two semitones higher per step.
+    If nRungs is negative, returns a lower pitch, and uses the
+    descending scale intervals instead of the ascending ones if they're
+    different.
     """
     if not isinstance(pitch, (int, float)):
         raise TypeError("Pitch must be a number.")
 
-    if not isinstance(nSteps, int):
-        raise TypeError("nSteps must be an integer.")
+    if not isinstance(nRungs, int):
+        raise TypeError("nRungs must be an integer.")
 
-    if nSteps == 0:
+    if nRungs == 0:
         return pitch
 
-    index = pianoIndex(pitch)
-    if (
-        index is None
-     or index + 2 > len(PIANO_KEYS) - 1
-     or index - 2 < 0
-    ):
-        index = None
+    # Compute how many octaves above or below our scale base we are
+    fundamental = currentFundamental()
 
-    if index is None:
-        return pitch * SEMITONE ** (2 * nSteps)
+    # Multiplier between the fundamental pitch and the given pitch
+    multiplier = pitch / fundamental
 
-    elif nSteps >= 0:
-        nextName, nextFreq = PIANO_KEYS[index + 1]
-        if 'b' in nextName: # it's a black key, so we skip it
-            nextFreq = PIANO_KEYS[index + 2][1]
+    # Compute number of half-steps above/below the fundamental, rounded
+    # to the nearest half-step.
+    halfSteps = round(math.log(multiplier) / math.log(SEMITONE))
 
-        if nSteps == 1:
-            return nextFreq
-        else:
-            return stepUpFrom(nextFreq, nSteps - 1)
+    # Get intervals for the current key
+    if nRungs > 0:
+        intervals = ascendingScaleIntervals()
+    else: # must be < 0, as == 0 was handled above
+        intervals = descendingScaleIntervals()
 
+    # Compute what counts as an octave (normally this should be 12)
+    octaveHalfSteps = sum(intervals)
+
+    # Compute which octave we're in
+    octave = halfSteps // octaveHalfSteps
+
+    # Compute how many steps along that octave we are
+    stepsInOctave = halfSteps % octaveHalfSteps
+
+    # Find nearest rung of the current key -- our starting rung
+    stepsSoFar = 0
+    stepsAboveFundamental = octave * octaveHalfSteps
+    for rung in range(len(intervals)):
+        intHere = intervals[rung]
+        if stepsSoFar + intHere > stepsInOctave:
+            below = stepsInOctave - stepsSoFar
+            above = stepsSoFar + intHere - stepsInOctave
+            if below < above:
+                stepsAboveFundamental += stepsSoFar
+                break
+            else:
+                stepsAboveFundamental += stepsSoFar + intHere
+                rung += 1
+                break
+        stepsSoFar += intHere
     else:
-        nextName, nextFreq = PIANO_KEYS[index - 1]
-        if 'b' in nextName: # it's a black key, so we skip it
-            nextFreq = PIANO_KEYS[index - 2][1]
+        # if we reach the end of the loop without breaking, we must have
+        # hit highest end of octave
+        stepsAboveFundamental += octaveHalfSteps
 
-        if nSteps == -1:
-            return nextFreq
-        else:
-            return stepUpFrom(nextFreq, nSteps + 1)
+    # Loop upwards through intervals to increase the pitch
+    if nRungs > 0:
+        for climb in range(rung, rung + nRungs):
+            stepsAboveFundamental += intervals[climb % len(intervals)]
+    else: # must be < 0, since == 0 was handled above
+        for climb in range(rung, rung + nRungs, -1):
+            stepsAboveFundamental -= intervals[(climb - 1) % len(intervals)]
+
+    return halfStepUpFrom(fundamental, stepsAboveFundamental)
 
 
-def stepDownFrom(pitch, nSteps=1):
+def climbDownFrom(pitch, nRungs=1):
     """
-    Returns the tone that's `nSteps` (default 1) whole-step(s) below the
-    given pitch, on an 8-note octave. Detects the nearest piano-key pitch
-    and gives the tone one step below that.
-
-    If the pitch is out of range of the piano keys, simply returns a
-    pitch two semitones lower.
+    The opposite of `climbUpFrom`. Equivalent to calling that function
+    with the same pitch and -nRungs as the argument.
     """
-    return stepUpFrom(pitch, -nSteps)
-
-
-def leapUpFrom(pitch, nSteps=1):
-    """
-    Works like `stepUpFrom`, but for the pentatonic scale. Note that the
-    pentatonic scale has much more limited range than the piano keys.
-    When given a pitch that falls between notes on the pentatonic scale,
-    this function finds the nearest pentatonic pitch and uses that as the
-    basis for the pitch.
-
-    When outside of the range of the pentatonic scale, this function
-    returns a note that's four semitones above (or below, if `nSteps` is
-    negative) the given pitch.
-    """
-    if not isinstance(pitch, (int, float)):
-        raise TypeError("Pitch must be a number.")
-
-    if not isinstance(nSteps, int):
-        raise TypeError("nSteps must be an integer.")
-
-    if nSteps == 0:
-        return pitch
-
-    index = pentatonicIndex(pitch)
-    default = pitch * SEMITONE**(4 * nSteps)
-    if (
-        index is None
-     or index + 1 > len(PENTATONIC_NOTES) - 1
-     or index - 1 < 0
-     or default < PENTATONIC_NOTES[0][1]
-     or default > PENTATONIC_NOTES[-1][1]
-    ):
-        return default
-
-    if nSteps == 1:
-        return PENTATONIC_NOTES[index + 1][1]
-    elif nSteps == -1:
-        return PENTATONIC_NOTES[index - 1][1]
-    elif nSteps > 1:
-        return leapUpFrom(PENTATONIC_NOTES[index + 1][1], nSteps - 1)
-    else: # less than -1
-        return leapUpFrom(PENTATONIC_NOTES[index - 1][1], nSteps + 1)
-
-
-def leapDownFrom(pitch, nSteps=1):
-    """
-    The opposite of `leapUp`.
-
-    When outside of the range of the pentatonic scale, this function
-    returns a note that's four semitones above the given pitch.
-    """
-    return leapUpFrom(pitch, -nSteps)
+    return climbUpFrom(pitch, -nRungs)
 
 
 #-----------------#
@@ -1065,7 +1224,35 @@ def currentDrum():
 CURRENT_PITCH = A3 # noqa F821
 """
 The pitch to use when `addNote` is called. Change using `setPitch`, or
-`stepUp` and related functions.
+`climbUp` and related functions.
+"""
+
+CURRENT_KEY = ("C", "Major")
+"""
+The key to use when moving pitches up or down using `climbUp` or
+`climbDown`. Includes the first note of the scale as a string, and then
+the type of scale (another string). Alternatively, a list of semitone
+differences or a pair of such lists can take the place of the scale type
+to use a custom scale. The `SCALE_TYPES` variable lists pre-defined scale
+types that you can use.
+
+TODO: More diverse scale types; non-semitone based tunings?
+"""
+
+SCALE_TYPES = {
+    "Major": [2, 2, 1, 2, 2, 2, 1],
+    "Minor-Natural": [2, 1, 2, 2, 1, 2, 2],
+    "Minor-Harmonic": [2, 1, 2, 2, 1, 3, 1],
+    "Minor-Melodic": ([2, 1, 2, 2, 2, 2, 1], [2, 2, 1, 2, 2, 1, 2]),
+    "Pentatonic-Major": [2, 2, 3, 2, 3],
+    "Pentatonic-Minor": [3, 2, 2, 3, 2],
+    "Pentatonic-Yo": [2, 3, 2, 2, 3],
+    "Pentatonic-In": [1, 4, 2, 3, 2],
+}
+"""
+A mapping from scale types to lists of semitone gaps between notes in
+that scale, or in some cases, pairs of lists for ascending and descending
+gap lists when these are different.
 """
 
 
@@ -1074,7 +1261,7 @@ def setPitch(pitch):
     Sets the current pitch value to the given pitch (an integer or
     floating-point number in Hertz). This pitch value will be used for
     subsequent calls to `addNote` until `setPitch` is called again or
-    some other pitch-modifying function is used (e.g., `stepUp`,
+    some other pitch-modifying function is used (e.g., `climbUp`,
     `halfStepDown`, etc.).
 
     You can use the constants defined in this module as pitch values,
@@ -1150,41 +1337,22 @@ def halfStepDown(nSteps=1):
     halfStepUp(-nSteps)
 
 
-def stepUp(nSteps=1):
+def climbUp(nRungs=1):
     """
-    Modifies the current pitch value to be `nSteps` full step(s) higher
-    than its old value. This uses `stepUpFrom` to compute the new pitch.
-    `nSteps` defaults to 1 if no argument is provided.
+    Modifies the current pitch value to be `nRungs` notes higher on the
+    current scale than the old value. Uses `climbUpFrom` to compute the
+    new pitch. `nRungs` defaults to 1 if no argument is provided.
     """
     global CURRENT_PITCH
-    CURRENT_PITCH = stepUpFrom(CURRENT_PITCH, nSteps)
+    CURRENT_PITCH = climbUpFrom(CURRENT_PITCH, nRungs)
 
 
-def stepDown(nSteps=1):
+def climbDown(nRungs=1):
     """
-    The opposite of `stepUp`: reduces the current pitch value. `nSteps`
+    The opposite of `climbUp`: reduces the current pitch value. `nRungs`
     defaults to 1.
     """
-    stepUp(-nSteps)
-
-
-def leapUp(nSteps=1):
-    """
-    Increases the current pitch value by `nSteps` pentatonic scale steps
-    (which are larger than the scientific notation steps used by `stepUp`
-    and `stepDown`). `leapUpFrom` is used to compute the new pitch value.
-    `nSteps` defaults to 1.
-    """
-    global CURRENT_PITCH
-    CURRENT_PITCH = leapUpFrom(CURRENT_PITCH, nSteps)
-
-
-def leapDown(nSteps=1):
-    """
-    The opposite of `leapUp`: recudes the current pitch value. `nSteps`
-    defaults to 1.
-    """
-    leapUp(-nSteps)
+    climbUp(-nRungs)
 
 
 CURRENT_TIME = 0
@@ -1603,7 +1771,7 @@ def addNote(duration):
     The specific sound used is determined by the current instrument,
     pitch, and volume values (see e.g., `currentInstrumentName`,
     `currentPitchName`, and `currentVolume`). Use `setInstrument`,
-    `setPitch`, and/or `setVolume` or related functions like `stepUp` or
+    `setPitch`, and/or `setVolume` or related functions like `climbUp` or
     `louder` to adjust these values before calling `addNote`.
 
     Note that `addSpecificNote` is used to do the heavy lifting here.
@@ -2143,6 +2311,40 @@ def test_name():
     assert True
 
 
+def test_scales():
+    """
+    A test function that produces a few different kinds of scales.
+    """
+    eraseTrack()
+    setInstrument(keyboard)
+
+    for f, st, n in [
+        ("C", "Major", 8),
+        ("Ab", "Minor-Natural", 8),
+        ("E", "Minor-Melodic", 8),
+        ("Fs", "Major", 8),
+        ("C", "Pentatonic-Major", 6),
+        ("A", "Pentatonic-Yo", 6),
+        ("A", "Pentatonic-In", 6),
+    ]:
+        setFundamental(f)
+        setScaleType(st)
+        setPitch(currentFundamental())
+        climbUp(n * 3)
+        p = currentPitch()
+        for i in range(n):
+            addNote(0.05)
+            climbUp()
+        for i in range(n):
+            climbDown()
+            addNote(0.05)
+        addRest(0.25)
+        assert currentPitch() == p
+
+    printTrack()
+    playTrack()
+
+
 def test_tune():
     """
     A test function that produces a drums-and-keyboard two-track tune,
@@ -2268,5 +2470,6 @@ def test_limit():
 if __name__ == "__main__":
     #test_name()
     #test_tune()
+    #test_scales()
     test_poly()
     #test_limit()
